from . import mpmatching

class Matcher:
    def __init__(self, G, trellis):
        self.G
        self.trellis

    def match(self):
        path_prob, predecessor = mpmatching.viterbi_search(self.G_interp, self.trellis, "start", "target")

        return results


import networkx as nx
import numpy as np
from pyproj import Geod
from shapely.geometry import Point, LineString

from pytrack.graph import utils

geod = Geod(ellps="WGS84")
EARTH_RADIUS_M = 6_371_009  # distance in meters


def get_bearing(lat1, lon1, lat2, lon2):
    """ Get bearing between two points.

    Parameters
    ----------
    lat1: float
        Latitude of the first point specified in decimal degrees
    lon1: float
        Longitude of the first point specified in decimal degrees
    lat2: float
        Latitude of the second point specified in decimal degrees
    lon2: float
        Longitude of the second point specified in decimal degrees

    Returns
    ----------
    bearing: float
        Bearing between two points.
    """

    bearing, _, _ = geod.inv(lon1, lat1, lon2, lat2)
    return bearing


def haversine_dist(lat1, lon1, lat2, lon2, earth_radius=EARTH_RADIUS_M):
    """ Calculate the great circle distance between two points on the earth (specified in decimal degrees)

    Parameters
    ----------
    lat1: float
        Latitude of the first point specified in decimal degrees
    lon1: float
        Longitude of the first point specified in decimal degrees
    lat2: float
        Latitude of the second point specified in decimal degrees
    lon2: float
        Longitude of the second point specified in decimal degrees

    earth_radius: float, optional, default: 6371009.0 meters
        Earth's radius
    Returns
    ----------
    dists: float
        Distance in units of earth_radius
    """
    # convert decimal degrees to radians 
    lon1, lat1, lon2, lat2 = map(np.deg2rad, [lon1, lat1, lon2, lat2])

    dlon = lon2 - lon1
    dlat = lat2 - lat1

    # haversine formula 
    h = np.sin(dlat / 2) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2) ** 2
    h = np.minimum(1, h)  # protect against floating point errors
    arc = 2 * np.arcsin(np.sqrt(h))

    dist = arc * earth_radius

    return dist


def enlarge_bbox(north, south, west, east, dist):
    """ Method that expands a bounding box by a specified distance.

    Parameters
    ----------
    north: float
        Northern latitude of bounding box.
    south: float
        Southern latitude of bounding box.
    west: float
        Western longitude of bounding box.
    east: float
        Eastern longitude of bounding box.

    dist: float
        Distance in meters indicating how much to expand the bounding box.

    Returns
    ----------
    north, south, west, east: float
        North, south, west, east coordinates of the expanded bounding box
    """
    delta_lat = (dist / EARTH_RADIUS_M) * (180 / np.pi)
    lat_mean = np.mean([north, south])
    delta_lng = (dist / EARTH_RADIUS_M) * (180 / np.pi) / np.cos(lat_mean * np.pi / 180)
    north = north + delta_lat
    south = south - delta_lat
    east = east + delta_lng
    west = west - delta_lng

    return north, south, west, east


def add_edge_lengths(G, precision=3):
    """ Method that adds the length of individual edges to the graph.

    Parameters
    ----------
    G: networkx.MultiDiGraph
        Road network graph.
    precision: float, optional, default: 3
        Number of decimal digits of the length of individual edges.
    Returns
    ----------
    G: networkx.MultiDiGraph
        Street network graph.
    """
    uvk = tuple(G.edges)

    lat = G.nodes(data='y')
    lon = G.nodes(data='x')

    lat_u, lon_u, lat_v, lon_v = list(zip(*[(lat[u], lon[u], lat[v], lon[v]) for u, v, _ in uvk]))

    dists = haversine_dist(lat_u, lon_u, lat_v, lon_v).round(precision)
    dists[np.isnan(dists)] = 0
    nx.set_edge_attributes(G, values=dict(zip(uvk, dists)), name='length')
    return G


def interpolate_graph(G, dist=1):
    """ Method that creates a graph by interpolating the nodes of a graph.

    Parameters
    ----------
    G: networkx.MultiDiGraph
        Road network graph.
    dist: float, optional, default: 1
        Distance between one node and the next.
    Returns
    ----------
    G: networkx.MultiDiGraph
        Street network graph.
    """
    G = G.copy()

    edges_toadd = []
    nodes_toadd = []

    for edge in list(G.edges(data=True)):
        u, v, data = edge
        # oneway = data["oneway"]
        geom = data["geometry"]
        data["length"] = dist

        G.remove_edge(u, v)

        XY = [xy for xy in _interpolate_geom(geom, dist=dist)]

        # generate nodes id
        uv = [(utils.get_unique_number(*u), utils.get_unique_number(*v)) for u, v in zip(XY[:-1], XY[1:])]
        # edges_interp = [LineString([u, v]) for u, v in zip(XY[:-1], XY[1:])]
        data_edges = [(lambda d: d.update({"geometry": LineString([u, v])}) or d)(data.copy()) for u, v in
                      zip(XY[:-1], XY[1:])]
        edges_toadd.extend([(*uv, data) for uv, data in zip(uv, data_edges)])

        nodes_toadd.extend([(utils.get_unique_number(lon, lat),
                             {"x": lon, "y": lat, "geometry": Point(lon, lat)}) for lon, lat in XY])

    G.add_edges_from(edges_toadd)
    G.add_nodes_from(nodes_toadd)

    G.remove_nodes_from(list(nx.isolates(G)))
    G.interpolation = True
    return G


def _interpolate_geom(geom, dist=1):
    """ Generator that interpolates a geometry created using the ``shapely.geometry.LineString`` method.

    Parameters
    ----------
    geom: shapely.geometry.LineString
        Geometry to be interpolated.
    dist: float, optional, default: 1
        Distance between one node and the next.
    Returns
    ----------
    ret: generator
        Interpolated geometry.
    """
    # TODO: use geospatial interpolation see: npts method in https://pyproj4.github.io/pyproj/stable/api/geod.html
    num_vert = max(round(geod.geometry_length(geom) / dist), 1)
    for n in range(num_vert + 1):
        point = geom.interpolate(n / num_vert, normalized=True)
        yield point.x, point.y


def interpolate_geom(geom, dist=1):
    """ Method that interpolates a geometry created using the ``shapely.geometry.LineString`` method.

    Parameters
    ----------
    geom: shapely.geometry.LineString
        Geometry to be interpolated.
    dist: float, optional, default: 1
        Distance between one node and the next.
    Returns
    ----------
    geom: shapely.geometry
        Interpolated geometry.
    """
    if isinstance(geom, LineString):
        return LineString([xy for xy in _interpolate_geom(geom, dist)])



# This lets you use package.module.Class as package.Class in your code.
from .visualization import Map

# This lets Sphinx know you want to document package.module.Class as package.Class.
__all__ = ['Map']


import json
import os
from pathlib import Path

import cv2
import requests

PREV_PAN_ID = None


def make_video(images, output_path, fourcc, fps=1, size=(640, 640), is_color=True):
    """ Makes video using xvid codec. Increase FPS for faster timelapse.

    Parameters
    ----------
    images: list
        List of the image paths to be added as frames to the video.
    output_path: str
        Path to output video, must end with .avi.
    fps: int, optional, default: 1
        Desired frame rate.
    size: tuple, optional, default: (640, 640)
        Size of the video frames.
    is_color: bool, optional, default: True
        If it is True, the encoder will expect and encode color frames, otherwise it will work with grayscale frames.

    :return: The function does not return anything. It directly saves the video at the position indicated in output_path.
    """
    vid = cv2.VideoWriter(output_path, fourcc, fps, size, is_color)
    for image in images:
        vid.write(cv2.imread(image))
    vid.release()
    cv2.destroyAllWindows()


def extract_streetview_pic(point, api_key, size="640x640", heading=90, pitch=-10):
    """ Extract street view pic.

    Parameters
    ----------
    point: tuple
        The lat/lng value of desired location.
    api_key: str
        Street View Static API key. It allows you to monitor your application's API usage in the Google Cloud Console,
        and ensures that Google can contact you about your application if necessary.
    size: str, optional, default: "640x640"
        Specifies the output size of the image in pixels.
    heading: int, optional, default: 90
        indicates the compass heading of the camera. Accepted values are from 0 to 360 (both values indicating North,
        with 90 indicating East, and 180 South). If no heading is specified, a value will be calculated that directs
        the camera towards the specified location, from the point at which the closest photograph was taken.
    pitch: int, optional, default: -10
        specifies the up or down angle of the camera relative to the Street View vehicle. This is often, but not always,
        flat horizontal. Positive values angle the camera up (with 90 degrees indicating straight up); negative values
        angle the camera down (with -90 indicating straight down).

    Returns
    -------
    pic: request.content
        Image get from Street View.
    meta: json
        Data about Street View panoramas.
    """

    global PREV_PAN_ID

    meta_base = 'https://maps.googleapis.com/maps/api/streetview/metadata?'
    pic_base = 'https://maps.googleapis.com/maps/api/streetview?'

    lat = point[0]
    lon = point[1]

    meta_params = {'key': api_key,
                   'location': f'{lat}, {lon}',
                   'source': 'outdoor'}

    pic_params = {
        'size': size,  # max 640x640 pixels
        'location': f'{lat}, {lon}',
        'heading': str(heading),
        'pitch': str(pitch),
        'key': api_key,
        'source': 'outdoor'
    }

    # Do the request and get the response data
    meta_response = requests.get(meta_base, params=meta_params)
    meta = meta_response.json()
    meta_response.close()
    if meta["status"] == "REQUEST_DENIED":
        print(meta)
    if (meta["status"] == "OK") and (meta["pano_id"] != PREV_PAN_ID):
        PREV_PAN_ID = meta["pano_id"]
        pic_response = requests.get(pic_base, params=pic_params)

        pic = pic_response.content
        pic_response.close()
    else:
        meta = None
        pic = None

    return pic, meta


def save_streetview(pic, meta, folder_path):
    """ Save streetview pic and metadata in the desired path.

    Parameters
    ----------
    pic: request.content
        Image get from Street View.
    meta: json
        Data about Street View panoramas.
    folder_path: str
        Desired path of the folder where save pic and metadata.

    :return: The function does not return anything. It directly saves the pic and metadata at the position indicated in folder_path.
    """
    Path(os.path.join(folder_path)).mkdir(parents=True, exist_ok=True)

    with open(os.path.join(folder_path, 'pic.png'), 'wb') as file:
        file.write(pic)

    with open(os.path.join(folder_path, 'metadata.json'), 'w+') as out_file:
        json.dump(meta, out_file)



import datetime as dt
from itertools import groupby

import networkx as nx
from shapely.geometry import Point, LineString

from . import distance
from . import download

useful_tags_way = [
    "bridge",
    "tunnel",
    "oneway",
    "lanes",
    "ref",
    "name",
    "highway",
    "maxspeed",
    "service",
    "access",
    "area",
    "landuse",
    "width",
    "est_width",
    "junction",
]


def graph_from_bbox(north, south, west, east, simplify=True, network_type='drive', buffer_dist=0):
    """ Create a graph from OpenStreetMap within some bounding box.

    Parameters
    ----------
    north: float
        Northern latitude of bounding box.
    south: float
        southern latitude of bounding box.
    west: float
        Western longitude of bounding box.
    east: float
        Eastern longitude of bounding box.
    simplify: bool, optional, default: True
        if True, simplify graph topology with the ``simplify_graph`` method.
    network_type: str, optional, default: 'drive'
        Type of street network to obtain.
    buffer_dist: float, optional, default: 0
        Distance in meters indicating how much to expand the bounding box.
    Returns
    -------
    G: networkx.MultiDiGraph
        Street network graph.
    """
    bbox_buffer = distance.enlarge_bbox(north, south, west, east, buffer_dist)
    response_json = download.osm_download(bbox_buffer, network_type=network_type)
    G = create_graph(response_json)

    if simplify:
        G.graph["simplified"] = True
        G = _simplification(G, response_json)
    else:
        G.graph["simplified"] = False
    return G


def _to_simplify(segment):
    """ Method that determines whether a segment of the graph must be simplified.

    Parameters
    ----------
    segment: list
        Graph segment to be evaluated. A segment is a list of node IDs.
    Returns
    -------
    ret: bool
        Whether or not to simplify a segment.
    """
    if len(segment) > 2:
        return True
    else:
        return False


def _simplification(G, response_json):
    """ Method that simplify a networkx graph.

    Parameters
    ----------
    G: networkx.MultiDiGraph
        Street network graph.
    response_json: json
        Response of OpenStreetMap API service got with``osm_download``method
    Returns
    -------
    G: networkx.MultiDiGraph
        Simplified street network graph.
    """
    _, paths = get_nodes_edges(response_json)
    for path in paths.values():
        is_oneway = _is_oneway(path, False)

        nodes = path.pop("nodes")

        junction = [False, False]
        if is_oneway:
            junction[1:1] = [True if G.degree[node] > 3 else False for node in nodes[1:-1]]
        else:
            junction[1:1] = [True if G.degree[node] > 4 else False for node in nodes[1:-1]]

        all_nodes_to_remove = []
        all_edges_to_add = []

        for segment in _split_at_values(nodes, junction):
            if _to_simplify(segment):
                all_edges_to_add.append(_build_edge(G, segment, path, is_oneway, reverse=False))

                if not is_oneway:
                    all_edges_to_add.append(_build_edge(G, segment, path, is_oneway, reverse=True))

                all_nodes_to_remove.extend(segment[1:-1])

        G.remove_nodes_from(set(all_nodes_to_remove))

        for edge in all_edges_to_add:
            G.add_edge(edge["origin"], edge["destination"], **edge["attr_dict"])

    return G


def _build_edge_attribute(G, segment, segment_attributes, is_oneway):
    """ Method that build the dictionary of attributes of a segment.

    Parameters
    ----------
    G: networkx.MultiDiGraph
        Street network graph.
    segment: list
        Graph segment to be evaluated. A segment is a list of node IDs.
    segment_attributes: dict
        Dictionary of road segment attributes.
    is_oneway: bool
        Indicates whether a street segment is oneway.
    Returns
    -------
    attribute: dict
        Dictionary of road segment attributes.
    """
    attribute = segment_attributes.copy()
    # attribute = {k: segment_attributes[k] for k in useful_tags_way if k in segment_attributes}

    attribute["geometry"] = LineString([Point((G.nodes[node]["x"],
                                               G.nodes[node]["y"])) for node in segment])
    attribute["oneway"] = is_oneway

    lat, lon = zip(*[(G.nodes[node]["y"], G.nodes[node]["x"]) for node in segment])
    edge_lengths = sum(distance.haversine_dist(lat[:-1], lon[:-1], lat[1:], lon[1:])).round(3)

    attribute["length"] = edge_lengths

    return attribute


def _build_edge(G, segment, segment_attributes, is_oneway, reverse=False):
    """ Method that builds an edge of the network street graph.

    Parameters
    ----------
    G: networkx.MultiDiGraph
        Street network graph.
    segment: list
        Graph segment to be evaluated. A segment is a list of node IDs.
    segment_attributes: dict
        Dictionary of road segment attributes.
    is_oneway: bool
        Indicates whether a street segment is oneway.
    reverse: bool, optional, default: False
        Indicates whether invert the direction of the edge.
    Returns
    -------
    edge: dict
        Edges to be added to the street network graph.
    """
    if not reverse:
        segment_attributes = _build_edge_attribute(G, segment, segment_attributes, is_oneway)

        edge = {"origin": segment[0], "destination": segment[-1],
                "attr_dict": segment_attributes}
        return edge
    else:
        segment_attributes = _build_edge_attribute(G, segment[::-1], segment_attributes, is_oneway)
        edge = {"origin": segment[-1], "destination": segment[0],
                "attr_dict": segment_attributes}
        return edge


def _split_at_values(nodes, junction):
    """ Method that creates a networkx.MultiDiGraph representing a network street graph.

    Parameters
    ----------
    nodes: list
        List of node IDs.
    junction: list
        It is a list of boolean conditions. If true, split the segment, otherwise not.
    Returns
    -------
    ret: generator
        Street network graph.
    """
    indices = [i for i, x in enumerate(nodes) if junction[i]]
    for start, end in zip([0, *indices], [*indices, len(nodes)]):
        yield nodes[start:end + 1]


def create_graph(response_json):
    """ Method that creates a networkx.MultiDiGraph representing a network street graph.

    Parameters
    ----------
    response_json: json
        Response of OpenStreetMap API service got with``osm_download``method.
    Returns
    -------
    G: networkx.MultiDiGraph
        Street network graph.
    """
    # TODO: nella creazione del grafo rimuovi intersezioni multiple, in modo tale che c'è un solo nodo intersezione
    #  per grafo create the graph as a MultiDiGraph and set its meta-attributes
    metadata = {
        'created_date': "{:%Y-%m-%d %H:%M:%S}".format(dt.datetime.now()),
        'created_with': f"PyTrack 1.0.0",
        'crs': "epsg:4326",
    }
    G = nx.MultiDiGraph(**metadata)

    nodes, paths = get_nodes_edges(response_json)

    # add each osm node to the graph
    for node, data in nodes.items():
        G.add_node(node, **data)

    add_edges(G, paths, bidirectional=False)

    # add length (haversine distance between nodes) attribute to each edge
    if len(G.edges) > 0:
        G = distance.add_edge_lengths(G)

    return G


def _oneway_path_values(path):
    """ Checks whether an OSM path is oneway.

    Parameters
    ----------
    path: dict
        Dictionary that describes an OSM path.
    Returns
    -------
    ret: dict
        Indicates whether an OSM path is oneway.
    """
    return {path[key] for key in path.keys() if key.startswith("oneway") and path[key] == "no"}


def _is_oneway(path, bidirectional):
    """ Checks whether an OSM path is oneway.

    Parameters
    ----------
    path: dict
        Dictionary that describes an OSM path.
    bidirectional: bool
        Indicates whether an edge is bidirectional.
    Returns
    -------
    is_oneway: bool
        Indicates whether an OSM path is oneway.
    """

    no_oneway_values = {"no", "false", "0", "reversible", "alternating"}

    oneway_path_values = _oneway_path_values(path)
    is_oneway = oneway_path_values.isdisjoint(no_oneway_values) and (
        not not oneway_path_values) and not bidirectional

    return is_oneway


def _is_reversed(path):
    """ Checks whether an OSM path is reversed.

    Parameters
    ----------
    path: dict
        Dictionary that describes an OSM path.
    Returns
    -------
    is_reversed: bool
        Indicates whether an OSM path is reversed.
    """
    reversed_values = {"-1", "reverse", "T"}

    oneway_path_values = _oneway_path_values(path)
    is_reversed = not oneway_path_values.isdisjoint(reversed_values)

    return is_reversed


def add_edges(G, paths, bidirectional=False, all_oneway=False):
    """ Add OSM edges to a ``networkx.MultiDiGraph``.

    Parameters
    ----------
    G: networkx.MultiDiGraph
        Street network graph.
    paths: dict
        Dictionary of OSM paths.
    bidirectional: bool, optional, default: False
        Indicates whether an edge is bidirectional.
    all_oneway: bool, optional, default: False
        Indicates whether an edge is oneway.
    """

    # TODO: "junction": "roundabout" è oneway
    for path in paths.values():
        nodes = path.pop('nodes')

        is_oneway = _is_oneway(path, bidirectional)
        is_reversed = _is_reversed(path)

        if is_oneway and is_reversed:
            nodes.reverse()

        if not all_oneway:
            path['oneway'] = is_oneway

        edges = list(zip(nodes[:-1], nodes[1:]))
        if not is_oneway:
            edges.extend([(v, u) for u, v in edges])

        G.add_edges_from(edges, **path)


def convert_edge(element):
    """ Convert an OSM edge into an edge to construct a street network graph.

    Parameters
    ----------
    element: dict
        An OSM path.
    Returns
    -------
    path: dict
        Dictionary for an OSM path.
    """
    # add OSM path id and remove consecutive duplicate nodes in the list of path's nodes
    path = {'osmid': element['id'], 'nodes': [group for group, _ in groupby(element['nodes'])]}

    # add tags
    path.update(element['tags'])

    return path


def convert_node(element):
    """ Convert an OSM node into a node to construct a street network graph.

    Parameters
    ----------
    element: dict
        An OSM node.
    Returns
    -------
    path: dict
        Dictionary for an OSM node.
    """

    # add node's GPS coordinates
    node = {'y': element['lat'], 'x': element['lon']}

    # add tags
    node.update(element['tags']) if "tags" in element else None

    return node


def get_nodes_edges(response_json):
    """ Extract nodes and paths from the OpenStreetMap query response.

    Parameters
    ----------
    response_json: json
        Response of OpenStreetMap API service got with``osm_download``method
    Returns
    -------
    nodes: dict
        Dictionary of OSM nodes.
    paths: dict
        Dictionary of OSM paths.
    """

    nodes = dict()
    paths = dict()

    for element in response_json['elements']:
        if element['type'] == "node":
            nodes[element['id']] = convert_node(element)
        elif element['type'] == "way":
            paths[element['id']] = convert_edge(element)

    return nodes, paths



from inspect import signature

import folium
import matplotlib.pyplot as plt
import networkx as nx
from shapely.geometry import LineString

from pytrack.graph import utils
from pytrack.matching import mpmatching_utils


class Map(folium.Map):
    """ This class extends the ``folium.Map`` to add functionality useful to represent graphs and road paths.

    Parameters
    ----------
    location: tuple or list, optional, default: None
        Latitude and Longitude of Map (Northing, Easting).
    width: int or percentage string, optional, default: '100%')
        Width of the map.
    height: int or percentage string, optional, default: '100%'
        Height of the map.
    tiles: str, optional, default: 'OpenStreetMap'
        Map tileset to use. Can choose from a list of built-in tiles,
        pass a custom URL or pass `None` to create a map without tiles.
        For more advanced tile layer options, use the `TileLayer` class.
    min_zoom: int, optional, default: 0
        Minimum allowed zoom level for the tile layer that is created.
    max_zoom: int, optional, default: 18
        Maximum allowed zoom level for the tile layer that is created.
    zoom_start: int, optional, default 10
        Initial zoom level for the map.
    attr: string, optional, default: None
        Map tile attribution; only required if passing custom tile URL.
    crs : str, optional, default: 'EPSG3857'
        Defines coordinate reference systems for projecting geographical points
        into pixel (screen) coordinates and back.
        You can use Leaflet's values :
        * EPSG3857 : The most common CRS for online maps, used by almost all
        free and commercial tile providers. Uses Spherical Mercator projection.
        Set in by default in Map's crs option.
        * EPSG4326 : A common CRS among GIS enthusiasts.
        Uses simple Equirectangular projection.
        * EPSG3395 : Rarely used by some commercial tile providers.
        Uses Elliptical Mercator projection.
        * Simple : A simple CRS that maps longitude and latitude into
        x and y directly. May be used for maps of flat surfaces
        (e.g. game maps). Note that the y axis should still be inverted
        (going from bottom to top).
    control_scale : bool, optional, default: False
        Whether to add a control scale on the map.
    prefer_canvas : bool, optional, default: False
        Forces Leaflet to use the Canvas back-end (if available) for
        vector layers instead of SVG. This can increase performance
        considerably in some cases (e.g. many thousands of circle
        markers on the map).
    no_touch : bool, optional, default: False
        Forces Leaflet to not use touch events even if it detects them.
    disable_3d : bool, optional, default: False
        Forces Leaflet to not use hardware-accelerated CSS 3D
        transforms for positioning (which may cause glitches in some
        rare environments) even if they're supported.
    zoom_control : bool, optional, default: True
        Display zoom controls on the map.
    **kwargs : keyword arguments, optional, default: no attributes
        Additional keyword arguments are passed to Leaflets Map class:
        https://leafletjs.com/reference-1.6.0.html#map

    Returns
    -------
    Folium Map Object

    Notes
    -----
    See https://github.com/python-visualization/folium/blob/551b2420150ab56b71dcf14c62e5f4b118caae32/folium/folium.py#L69
    for a more detailed description

    """

    def __init__(
            self,
            location=None,
            width='100%',
            height='100%',
            left='0%',
            top='0%',
            position='relative',
            tiles='CartoDB positron',
            attr=None,
            min_zoom=0,
            max_zoom=18,
            zoom_start=15,
            min_lat=-90,
            max_lat=90,
            min_lon=-180,
            max_lon=180,
            max_bounds=False,
            crs='EPSG3857',
            control_scale=False,
            prefer_canvas=False,
            no_touch=False,
            disable_3d=False,
            png_enabled=False,
            zoom_control=True,
            **kwargs
    ):
        super().__init__(location, width, height, left, top, position, tiles, attr, min_zoom, max_zoom, zoom_start,
                         min_lat, max_lat, min_lon, max_lon, max_bounds, crs, control_scale, prefer_canvas, no_touch,
                         disable_3d, png_enabled, zoom_control, **kwargs)
        self.tiles = tiles
        folium.LatLngPopup().add_to(self)

    def _render_reset(self):
        for key in list(self._children.keys()):
            if key.startswith('cartodbpositron') or key.startswith('lat_lng_popup'):
                self._children.pop(key)
        children = self._children
        self.__init__(self.location, tiles=self.tiles)
        self.options = self.options
        for k, v in children.items():
            self.add_child(v)

    def add_graph(self, G, plot_nodes=False, edge_color="#3388ff", edge_width=3,
                  edge_opacity=1, radius=1.7, node_color="red", fill=True, fill_color=None,
                  fill_opacity=1):
        """ Add the road network graph created with ``pytrack.graph.graph.graph_from_bbox`` method

        Parameters
        ----------
        G: networkx.MultiDiGraph
            Road network graph.
        plot_nodes: bool, optional, default: False
            If true, it will show the vertices of the graph.
        edge_color: str, optional, default: "#3388ff"
            Colour of graph edges
        edge_width: float, optional, default: 3
            Width of graph edges
        edge_opacity: float, optional, default: 1
            Opacity of graph edges
        radius: float, optional, default: 1.7
            Radius of graph vertices
        node_color: str, optional, default: "red"
            Colour of graph vertices
        fill: bool, optional, default: True
            Whether to fill the nodes with color. Set it to false to disable filling on the nodes.
        fill_color: str or NoneType, default: None
            Fill color. Defaults to the value of the color option.
        fill_opacity: float, optional, default: 1
            Fill opacity
        """
        edge_attr = dict()
        edge_attr["color"] = edge_color
        edge_attr["weight"] = edge_width
        edge_attr["opacity"] = edge_opacity

        node_attr = dict()
        node_attr["color"] = node_color
        node_attr["fill"] = fill

        if not fill_color:
            node_attr["fill_color"] = fill_color
        else:
            node_attr["fill_color"] = fill_color
        node_attr["fill_opacity"] = fill_opacity

        nodes, edges = utils.graph_to_gdfs(G)
        edges = utils.graph_to_gdfs(G, nodes=False)

        fg_graph = folium.FeatureGroup(name='Graph edges', show=True)
        self.add_child(fg_graph)

        for geom in edges.geometry:
            edge = [(lat, lng) for lng, lat in geom.coords]
            folium.PolyLine(locations=edge, **edge_attr, ).add_to(fg_graph)

        if plot_nodes:
            fg_point = folium.FeatureGroup(name='Graph vertices', show=True)
            self.add_child(fg_point)
            for point, osmid in zip(nodes.geometry, nodes.osmid):
                folium.Circle(location=(point.y, point.x), popup=f"osmid: {osmid}", radius=radius, **node_attr).add_to(
                    fg_point)

        folium.LayerControl().add_to(self)

    def draw_candidates(self, candidates, radius):
        """ Draw the candidate nodes of the HMM matcher

        Parameters
        ----------
        candidates: dict
            Candidates' dictionary computed via ``pytrack.matching.candidate.get_candidates`` method
        radius: float
            Candidate search radius
        """
        fg_cands = folium.FeatureGroup(name='Candidates', show=True, control=True)
        fg_gps = folium.FeatureGroup(name="Actual GPS points", show=True, control=True)
        self.add_child(fg_cands)
        self.add_child(fg_gps)

        for i, obs in enumerate(candidates.keys()):
            folium.Circle(location=candidates[obs]["observation"], radius=radius, weight=1, color="black", fill=True,
                          fill_opacity=0.2).add_to(fg_gps)
            popup = f'{i}-th point \n Latitude: {candidates[obs]["observation"][0]}\n Longitude: ' \
                    f'{candidates[obs]["observation"][1]}'
            folium.Circle(location=candidates[obs]["observation"], popup=popup, radius=1, color="black",
                          fill=True, fill_opacity=1).add_to(fg_gps)

            # plot candidates
            for cand, label, cand_type in zip(candidates[obs]["candidates"], candidates[obs]["edge_osmid"],
                                              candidates[obs]["candidate_type"]):
                popup = f"coord: {cand} \n edge_osmid: {label}"
                if cand_type:
                    folium.Circle(location=cand, popup=popup, radius=2, color="yellow", fill=True,
                                  fill_opacity=1).add_to(
                        fg_cands)
                else:
                    folium.Circle(location=cand, popup=popup, radius=1, color="red", fill=True, fill_opacity=1).add_to(
                        fg_cands)

        del self._children[next(k for k in self._children.keys() if k.startswith('layer_control'))]
        self.add_child(folium.LayerControl())
        self._render_reset()

    def draw_path(self, G, trellis, predecessor, path_name="Matched path"):
        """ Draw the map-matched path

        Parameters
        ----------
        G: networkx.MultiDiGraph
            Road network graph.
        trellis: nx.DiGraph
            Trellis DAG graph created with ``pytrack.matching.mpmatching_utils.create_trellis`` method
        predecessor: dict
            Predecessors' dictionary computed with ``pytrack.matching.mpmatching.viterbi_search`` method
        path_name: str
            Name of the path to be drawn
        """

        fg_matched = folium.FeatureGroup(name=path_name, show=True, control=True)
        self.add_child(fg_matched)

        path_elab = mpmatching_utils.create_path(G, trellis, predecessor)

        edge_attr = dict()
        edge_attr["color"] = "green"
        edge_attr["weight"] = 4
        edge_attr["opacity"] = 1

        edge = [(lat, lng) for lng, lat in LineString([G.nodes[node]["geometry"] for node in path_elab]).coords]
        folium.PolyLine(locations=edge, **edge_attr, ).add_to(fg_matched)

        del self._children[next(k for k in self._children.keys() if k.startswith('layer_control'))]
        self.add_child(folium.LayerControl())
        self._render_reset()


def draw_trellis(T, figsize=None, dpi=None, node_size=500, font_size=8, **kwargs):
    """ Draw a trellis graph

    Parameters
    ----------
    T: networkx.DiGraph
        A directed acyclic graph
    figsize: (float, float), optional, default: [15.0, 12.0]
        Width, height figure size tuple in inches, optional
    dpi: float, optional, default: 300.0
        The resolution of the figure in dots-per-inch
    node_size: scalar or array, optional, default: 500
        Size of nodes.  If an array is specified it must be the same length as nodelist.
    font_size: int, optional, default: 8
        Font size for text labels
    kwargs: keyword arguments, optional, default: no attributes
        See networkx.draw_networkx_nodes(), networkx.draw_networkx_edges(),
        networkx.draw_networkx_labels() and matplotlib.pyplot.figure() for a description of optional keywords.

    Returns
    -------
    trellis_diag: matplotlib.pyplot.Figure
        Graphical illustration of the Trellis diagram used in the Hidden Markov Model process to find the path that best
        matches the actual GPS data
    """

    valid_node_kwargs = signature(nx.draw_networkx_nodes).parameters.keys()
    valid_edge_kwargs = signature(nx.draw_networkx_edges).parameters.keys()
    valid_label_kwargs = signature(nx.draw_networkx_labels).parameters.keys()
    valid_plt_kwargs = signature(plt.figure).parameters.keys()

    valid_nx_kwargs = (valid_node_kwargs | valid_edge_kwargs | valid_label_kwargs)

    # Create a set with all valid keywords across the three functions and
    # remove the arguments of this function (draw_networkx)
    valid_kwargs = (valid_nx_kwargs | valid_plt_kwargs) - {
        "G",
        "figsize",
        "dpi",
        "pos",
        "node_size",
        "font_size",
    }

    if any([k not in valid_kwargs for k in kwargs]):
        invalid_args = ", ".join([k for k in kwargs if k not in valid_kwargs])
        raise ValueError(f"Received invalid argument(s): {invalid_args}")

    nx_kwargs = {k: v for k, v in kwargs.items() if k in valid_nx_kwargs}
    plt_kwargs = {k: v for k, v in kwargs.items() if k in valid_plt_kwargs}

    if figsize is None:
        figsize = (15, 12)
    if dpi is None:
        dpi = 300

    plt.figure(figsize=figsize, dpi=dpi, **plt_kwargs)

    pos = nx.drawing.nx_pydot.graphviz_layout(T, prog='dot', root='start')
    trellis_diag = nx.draw_networkx(T, pos, node_size=node_size, font_size=font_size, **nx_kwargs)

    return trellis_diag



from collections import deque

from . import mpmatching_utils


def viterbi_search(G, trellis, start="start", target="target", beta=mpmatching_utils.BETA,
                   sigma=mpmatching_utils.SIGMA_Z):
    """ Function to compute viterbi search and perform Hidden-Markov Model map-matching.

    Parameters
    ----------
    G: networkx.MultiDiGraph
        Street network graph.
    trellis:
    start: str, optional, default: "start"
        Starting node.
    target: str, optional, default: "target"
        Target node.
    beta: float
        This describes the difference between route distances and great circle distances. See https://www.ismll.uni-hildesheim.de/lehre/semSpatial-10s/script/6.pdf
        for a more detailed description of its calculation.

    sigma: float
        It is an estimate of the magnitude of the GPS error. See https://www.ismll.uni-hildesheim.de/lehre/semSpatial-10s/script/6.pdf
        for a more detailed description of its calculation.

    Returns
    -------
    joint_prob: dict
        Joint probability for each node.
    predecessor: dict
        Predecessor for each node.

    Notes
    -----
    See https://www.ismll.uni-hildesheim.de/lehre/semSpatial-10s/script/6.pdf for a more detailed description of this
    method.
    """

    # Initialize joint probability for each node
    joint_prob = {}
    for u_name in trellis.nodes():
        joint_prob[u_name] = 0
    predecessor = {}
    queue = deque()

    queue.append(start)
    joint_prob[start] = mpmatching_utils.emission_prob(trellis.nodes[start]["candidate"], sigma)
    predecessor[start] = None

    while queue:
        # Extract node u
        u_name = queue.popleft()
        u = trellis.nodes[u_name]["candidate"]

        if u_name == target:
            break
        for v_name in trellis.successors(u_name):
            v = trellis.nodes[v_name]["candidate"]

            new_prob = joint_prob[u_name] * mpmatching_utils.transition_prob(G, u, v, beta) * \
                mpmatching_utils.emission_prob(v, sigma)

            if joint_prob[v_name] < new_prob:
                joint_prob[v_name] = new_prob
                predecessor[v_name.split("_")[0]] = u_name
            if v_name not in queue:
                queue.append(v_name)

    return joint_prob[target], predecessor



import copy

import numpy as np
import pandas as pd
from sklearn.neighbors import BallTree

from pytrack.graph import distance, utils


class Candidate:
    """ Class to represent a candidate element.

    Parameters
    ----------
    node_id: str
        OSM node ID.
    edge_osmid: str
        OSM edge ID.
    obs: tuple
        Coordinate of actual GPS points.
    great_dist: float
        Distance between candidate and actual GPS point.
    coord: tuple
        Candidate coordinate.
    Returns
    -------
    Candidate object
    """
    __slots__ = ["node_id", "edge_osmid", "obs", "great_dist", "coord"]

    def __init__(self, node_id, edge_osmid, obs, great_dist, coord):
        self.node_id = node_id
        self.edge_osmid = edge_osmid
        self.obs = obs
        self.great_dist = great_dist
        self.coord = coord


def get_candidates(G, points, interp_dist=1, closest=True, radius=10):
    """ Extract candidate points for Hidden-Markov Model map-matching approach.

    Parameters
    ----------
    G: networkx.MultiDiGraph
        Street network graph.
    points: list
        The actual GPS points.
    interp_dist: float, optional, default: 1
        Step to interpolate the graph. The smaller the interp_dist, the greater the precision and the longer
        the computational time.
    closest: bool, optional, default: True
        If true, only the closest point is considered for each edge.
    radius: float, optional, default: 10
        Radius of the search circle.
    Returns
    -------
    G: networkx.MultiDiGraph
        Street network graph.
    results: dict
        Results to be used for the map-matching algorithm.
    """

    G = G.copy()

    if interp_dist:
        G = distance.interpolate_graph(G, dist=interp_dist)

    geoms = utils.graph_to_gdfs(G, nodes=False).set_index(["u", "v"])[["osmid", "geometry"]]

    uv_xy = [[u, osmid, np.deg2rad(xy)] for uv, geom, osmid in
             zip(geoms.index, geoms.geometry.values, geoms.osmid.values)
             for u, xy in zip(uv, geom.coords[:])]

    index, osmid, xy = zip(*uv_xy)

    nodes = pd.DataFrame(xy, index=osmid, columns=["x", "y"])[["y", "x"]]

    ball = BallTree(nodes, metric='haversine')

    idxs, dists = ball.query_radius(np.deg2rad(points), radius / distance.EARTH_RADIUS_M, return_distance=True)
    dists = dists * distance.EARTH_RADIUS_M  # radians to meters

    if closest:
        results = dict()
        for i, (point, idx, dist) in enumerate(zip(points, idxs, dists)):
            df = pd.DataFrame({"osmid": list(np.array(index)[idx]),
                               "edge_osmid": list(nodes.index[idx]),
                               "coords": tuple(map(tuple, np.rad2deg(nodes.values[idx]))),
                               "dist": dist})

            df = df.loc[df.groupby('edge_osmid')['dist'].idxmin()].reset_index(drop=True)

            results[i] = {"observation": point,
                          "osmid": list(df["osmid"]),
                          "edge_osmid": list(df["edge_osmid"]),
                          "candidates": list(df["coords"]),
                          "candidate_type": np.full(len(df["coords"]), False),
                          "dists": list(df["dist"])}

    else:
        results = {i: {"observation": point,
                       "osmid": list(np.array(index)[idx]),
                       "edge_osmid": list(nodes.index[idx]),
                       "candidates": list(map(tuple, np.rad2deg(nodes.values[idx]))),
                       "candidate_type": np.full(len(nodes.index[idx]), False),
                       "dists": list(dist)} for i, (point, idx, dist) in enumerate(zip(points, idxs, dists))}

    return G, results


def elab_candidate_results(results, predecessor):
    """ Elaborate results of ``candidate.get_candidates`` method. It selects which candidate best matches the actual
    GPS points.

    Parameters
    ----------
    results: dict
        Output of ``candidate.get_candidates`` method.
    predecessor: dict
        Output of ``mpmatching.viterbi_search`` method.
    Returns
    -------
    results: dict
        Elaborated results.
    """
    results = copy.deepcopy(results)
    for key in list(results.keys())[:-1]:
        win_cand_idx = int(predecessor[str(key + 1)].split("_")[1])
        results[key]["candidate_type"][win_cand_idx] = True

    win_cand_idx = int(predecessor["target"].split("_")[1])
    results[list(results.keys())[-1]]["candidate_type"][win_cand_idx] = True
    return results



from decimal import Decimal

import pandas as pd
from shapely.geometry import LineString, Point


def get_unique_number(lon, lat):
    """ Assigns a unique identifier to a geographical coordinate.
    Parameters
    ----------
    lon: float
        Longitude of the point
    lat: float
        Latitude of the point
    Returns
    -------
    val: float
        Unique identifier.
    """
    if isinstance(lat, str):
        lat_double = float(lat)
    else:
        lat_double = lat
    if isinstance(lon, str):
        lon_double = float(lon)
    else:
        lon_double = lon

    lat_int = int((lat_double * 10 ** abs(Decimal(str(lat_double)).as_tuple().exponent)))
    lon_int = int((lon_double * 10 ** abs(Decimal(str(lon_double)).as_tuple().exponent)))

    val = abs((lat_int << 16 & 0xffff0000) | (lon_int & 0x0000ffff))
    val = val % 2147483647

    # alternative use hash or other
    # return int(str(lat_int)+str(lon_int))
    return val


def graph_to_gdfs(G, nodes=True, edges=True, node_geometry=True, edge_geometry=True):
    """ Convert a networkx.MultiDiGraph to node and/or edge pandas DataFrame.

    Parameters
    ----------
    G: networkx.MultiDiGraph
        Street network graph.
    nodes: bool, optional, default: True
        Whether to extract graph nodes.
    edges: bool, optional, default: True
        Whether to extract graph edges.
    node_geometry: bool, optional, default: True
        Whether to compute graph node geometries.
    edge_geometry: bool, optional, default: True
        Whether to extract graph edge geometries.
    Returns
    -------
    gdf_nodes: pandas.DataFrame
        Dataframe collecting graph nodes.
    gdf_edges: pandas.DataFrame
        Dataframe collecting graph edges
    """
    crs = G.graph["crs"]

    if nodes:
        nodes, data = zip(*G.nodes(data=True))

        if node_geometry:
            # convert node x/y attributes to Points for geometry column
            for d in data:
                d["geometry"] = Point(d["x"], d["y"])
            gdf_nodes = pd.DataFrame(data)
            gdf_nodes.insert(loc=0, column='osmid', value=nodes)
        else:
            gdf_nodes = pd.DataFrame(data)
            gdf_nodes.insert(loc=0, column='osmid', value=nodes)

    if edges:
        u, v, k, data = zip(*G.edges(keys=True, data=True))
        if edge_geometry:
            longs = G.nodes(data="x")
            lats = G.nodes(data="y")

            for d, src, tgt in zip(data, u, v):
                if "geometry" not in d:
                    d["geometry"] = LineString((Point((longs[src], lats[src])),
                                                Point((longs[tgt], lats[tgt]))))
            gdf_edges = pd.DataFrame(data)

        else:
            gdf_edges = pd.DataFrame(data)
            gdf_edges["geometry"] = None
            gdf_edges.crs = crs

        gdf_edges["u"], gdf_edges["v"], gdf_edges["key"] = u, v, k

        gdf_edges = gdf_edges[["u", "v", "key"] + gdf_edges.columns.to_list()[:-3]]

    if nodes and edges:
        return gdf_nodes, gdf_edges
    elif nodes:
        return gdf_nodes
    elif edges:
        return gdf_edges



import itertools
import math

import networkx as nx
from shapely.geometry import LineString

from pytrack.graph import distance
from pytrack.matching import candidate

SIGMA_Z = 4.07
BETA = 20


# A gaussian distribution
def emission_prob(u, sigma=SIGMA_Z):
    """ Compute emission probability of a node

    Parameters
    ----------
    u: dict
        Node of the graph.
    sigma: float
        It is an estimate of the magnitude of the GPS error. See https://www.ismll.uni-hildesheim.de/lehre/semSpatial-10s/script/6.pdf
        for a more detailed description of its calculation.

    Returns
    -------
    ret: float
        Emission probability of a node.
    """
    if u.great_dist:
        c = 1 / (sigma * math.sqrt(2 * math.pi))
        return c * math.exp(-(u.great_dist / sigma) ** 2)
    else:
        return 1


# A empirical distribution
def transition_prob(G, u, v, beta=BETA):
    """ Compute transition probability between node u and v.

    Parameters
    ----------
    G: networkx.MultiDiGraph
        Road network graph.
    u: dict
        Starting node of the graph.
    v: dict
        Target node of the graph.
    beta: float
        This describes the difference between route distances and great circle distances. See https://www.ismll.uni-hildesheim.de/lehre/semSpatial-10s/script/6.pdf
        for a more detailed description of its calculation.

    Returns
    -------
    ret: float
        Transition probability between node u and v.
    """
    c = 1 / BETA

    if u.great_dist and v.great_dist:
        delta = abs(
            nx.shortest_path_length(G, u.node_id, v.node_id, weight="length", method='dijkstra')
            - distance.haversine_dist(*u.coord, *v.coord))
        return c * math.exp(-delta / BETA) * 1e5
    else:
        return 1


def create_trellis(results):
    """ Create a Trellis graph.

    Parameters
    ----------
    results: dict
        Output of ``candidate.get_candidates`` method.
    Returns
    -------
    G: networkx.DiGraph
        A directed acyclic Trellis graph.
    """

    G = nx.DiGraph()

    prev_nodes = ["start"]
    G.add_node("start", candidate=candidate.Candidate("start", None, None, None, None))
    G.add_node("target", candidate=candidate.Candidate("start", None, None, None, None))

    for idx, item in results.items():
        obs = item["observation"]
        nodes, data_node = zip(
            *[(f"{idx}_{j}", {"candidate": candidate.Candidate(node_id, edge_osmid, obs, dist, cand)})
              for j, (node_id, edge_osmid, dist, cand) in
              enumerate(zip(item["osmid"], item["edge_osmid"], item["dists"], item["candidates"]))])
        edges = [(u, v) for u in prev_nodes for v in nodes]

        G.add_nodes_from(zip(nodes, data_node))
        G.add_edges_from(edges)
        prev_nodes = nodes

    edges = [(u, "target") for u in prev_nodes]
    G.add_edges_from(edges)
    return G


def create_path(G, trellis, predecessor):
    """ Create the path that best matches the actual GPS data.

    Parameters
    ----------
    G: networkx.MultiDiGraph
        Road network graph.
    trellis: networkx.DiGraph
        A directed acyclic graph.
    predecessor: dict
        Predecessor for each node.
    Returns
    -------
    path_elab: list
        list of node IDs.
    """
    u, v = list(zip(*[(u, v) for v, u in predecessor.items()][2:]))
    path = [(u, v) for u, v in zip(u, u[1:])]

    path_elab = [node for u, v in path for node in nx.shortest_path(G, trellis.nodes[u]["candidate"].node_id,
                                                                    trellis.nodes[v]["candidate"].node_id,
                                                                    weight='length')]
    path_elab = [k for k, g in itertools.groupby(path_elab)]
    return path_elab


def create_matched_path(G, trellis, predecessor):
    """ Create the path that best matches the actual GPS points. Route created based on results obtained from ``pmatching_utils.viterbi_search`` and ``mpmatching_utils.create_trellis`` methods.

    Parameters
    ----------
    G: networkx.MultiDiGraph
        Street network graph used to create trellis graph.
    trellis: networkx.DiGraph
        A directed acyclic Trellis graph.
    predecessor: dict
        Predecessor for each node.

    Returns
    -------
    node_ids: list
        List of ids of the nodes that compose the path.
    path_coords: list
        List of nodes' coordinates, in the form of tuple (lat, lon), composing the path.
    """
    node_ids = create_path(G, trellis, predecessor)
    path_coords = [(lat, lng) for lng, lat in LineString([G.nodes[node]["geometry"] for node in node_ids]).coords]
    return node_ids, path_coords



import requests


def get_filters(network_type='drive'):
    """ Get the filters with which to interrogate the OpenStreetMao API service.

    Parameters
    ----------
    network_type: str, optional, default: 'drive'
        Type of street network to obtain.
    Returns
    -------
    osm_filters: str
        Filters identifying the type of road network to be obtained
    """
    osm_filters = dict()

    osm_filters['drive'] = ('["highway"]["area"!~"yes"]["access"!~"private"]'
                            '["highway"!~"abandoned|bridleway|bus_guideway|construction|corridor|cycleway|'
                            'elevator|escalator|footway|path|pedestrian|planned|platform|proposed|raceway|steps|track"]'
                            '["service"!~"emergency_access|parking|parking_aisle|private"]')

    return osm_filters[network_type]


def osm_download(bbox, network_type=None):
    """ Get the OpenStreetMap response.

    Parameters
    ----------
    bbox: tuple
        bounding box within N, S, E, W coordinates.
    network_type: str or None, optional, default: None
        Type of street network to obtain.
    Returns
    -------
    response: json
        Response of the OpenStreetMao API service.
    """
    # TODO: add network_type statement
    north, south, west, east = bbox

    osm_filters = get_filters(network_type='drive')

    url_endpoint = 'https://maps.mail.ru/osm/tools/overpass/api/interpreter'

    timeout = 180
    out_resp = 'json'

    overpass_settings = f'[out:{out_resp}][timeout:{timeout}]'

    query_str = f'{overpass_settings};(way{osm_filters}({south}, {west}, {north}, {east});>;);out;'

    response_json = requests.post(url_endpoint, data={'data': query_str})

    size_kb = len(response_json.content) / 1000
    print(f'Downloaded {size_kb:,.2f}kB')

    return response_json.json()



import cv2


def make_video(images, output_path, fps=1, size=(640, 640), is_color=True):
    """ Makes video using xvid codec. Increase FPS for faster timelapse.

    Parameters
    ----------
    images: list
        List of the image paths to be added as frames to the video.
    output_path: str
        Path to output video, must end with .avi.
    fps: int, optional, default: 1
        Desired frame rate.
    size: tuple, optional, default: (640, 640)
        Size of the video frames.
    is_color: bool, optional, default: True
        If it is True, the encoder will expect and encode color frames, otherwise it will work with grayscale frames.

    :return: The function does not return anything. It directly saves the video at the position indicated in output_path.
    """
    fourcc = cv2.VideoWriter_fourcc(*"XVID")
    vid = cv2.VideoWriter(output_path, fourcc, fps, size, is_color)
    for image in images:
        vid.write(cv2.imread(image))
    vid.release()
    cv2.destroyAllWindows()



import setuptools
import os


def strip_comments(l):
    return l.split('#', 1)[0].strip()


def reqs(*f):
    return list(filter(None, [strip_comments(l) for l in open(
        os.path.join(os.getcwd(), *f)).readlines()]))


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='PyTrack-lib',
    version='2.0.0',
    packages=setuptools.find_packages(),
    # namespace_packages=['pytrack'],
    url='https://github.com/cosbidev/PyTrack',
    license='BSD-3-Clause-Clear',
    author='Matteo Tortora',
    author_email='m.tortora@unicampus.it',
    description='a Map-Matching-based Python Toolbox for Vehicle Trajectory Reconstruction',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: GIS",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Information Analysis"
    ],
    install_requires=reqs('requirements.txt')
)
